#
# Copyright (C) 2021 Associated Universities, Inc. Washington DC, USA.
#
# This file is part of NRAO Workspaces.
#
# Workspaces is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Workspaces is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Workspaces.  If not, see <https://www.gnu.org/licenses/>.
""" Pieces of any ingestion manifest """

import abc
import re
from enum import Enum
from pathlib import Path
from typing import Dict, List, Union

from ingest_envoy.schema import AbstractTextFile
from ingest_envoy.utilities import AncillaryProductType, ScienceProductType, Telescope

MANIFEST_FILENAME = "ingestion-manifest.json"
CURATOR_MANIFEST_FILENAME = "manifest.json"
INGESTION_ARTIFACTS_NAME = "ingestion_artifacts_"
TARFILE_EXT = ".tar"
WEBLOG_FILENAME = "weblog.tgz"
INIT_WEBLOG_FILENAME = "initial_weblog.tgz"
REIMAGING_FILENAME = "reimaging_resources.tgz"
PARAMETER_FILENAME = "SEIP_parameter.list"
SCIENCE_PRODUCT_PATTERN = re.compile("[a-zA-Z0-9._\\-+]*\\.tar")
JSON = Union[int, float, str, List["JSON"], Dict[str, "JSON"]]


class IngestionManifestKey(Enum):
    """Sections we expect to see in a manifest"""

    PARAMETERS = "parameters"
    INPUT_GROUP = "input_group"
    OUTPUT_GROUP = "output_group"
    SCIENCE_PRODUCTS = "science_products"
    ANCILLARY_PRODUCTS = "ancillary_products"
    REINGEST = "reingest"


class ParamsKey(Enum):
    """Keys in ManifestParameters"""

    PARAMETERS = "parameters"
    TELESCOPE = "telescope"
    REINGEST = "reingest"
    NGAS_INGEST = "ngas_ingest"
    CALIBRATE = "calibrate"
    INGESTION_PATH = "ingestion_path"
    ADDITIONAL_METADATA = "additional_metadata"
    COLLECTION_METADATA = "additional_metadata"  # needed for VLASS, realfast, elwa, alfalfa. Not yet implemented


class ReingestKey(Enum):
    """Keys in ReingestGroup"""

    TARGETS = "targets"
    LOCATOR = "locator"
    TYPE = "type"


class ManifestComponentIF(abc.ABC):
    """Interface for components of an ingestion manifest (including the manifest itself)"""

    @abc.abstractmethod
    def __eq__(self, other):
        return NotImplemented

    @abc.abstractmethod
    def to_dict(self) -> Dict:
        """
        Build a JSON-ifyable dictionary

        :return: json.dumps()-able dictionary object
        """


class InputScienceProduct(ManifestComponentIF):
    """Simplest type of science product: has only a locator"""

    def __init__(self, locator: str):
        self.locator = locator

    def __eq__(self, other):
        if isinstance(other, InputScienceProduct):
            return other.locator == self.locator

    def to_dict(self) -> JSON:
        """
        Turn me into a json-ifiable dict

        :return: dicty-me
        """
        return {"locator": self.locator}


class InputGroup(ManifestComponentIF):
    """Generic ingestion manifest input group"""

    def __init__(self, science_products: List[InputScienceProduct]):
        self.science_products = science_products

    def __eq__(self, other):
        if isinstance(other, InputGroup):
            return other.science_products == self.science_products

        return False

    def to_dict(self) -> JSON:
        """
        Turn me into a json-ifiable dict

        :return: dicty-me
        """
        me_dict = dict(self.__dict__)

        sps = me_dict[IngestionManifestKey.SCIENCE_PRODUCTS.value]
        sps = [sp.to_dict() for sp in sps]
        if len(sps) == 0:
            return {}

        to_return = {IngestionManifestKey.SCIENCE_PRODUCTS.value: sps}
        return to_return


class ManifestParameters(ManifestComponentIF):
    """Represents "parameters" section of ingestion manifest.
    ASSUMPTIONS:
    * EVLA CAL manifest has no "calibrate" parameter
    * "ngas_ingest" is always True (per our testing examples)
    """

    def __init__(
        self,
        telescope: Telescope,
        ngas_ingest: bool,
        staging_source_dir: Path | None,
        additional_metadata: AbstractTextFile = None,
        calibrate: bool = None,
    ):
        self.telescope = telescope

        self.ngas_ingest = ngas_ingest
        if calibrate is not None:
            self.calibrate = calibrate

        if staging_source_dir is not None:
            self.staging_source_dir = staging_source_dir
        self.additional_metadata = additional_metadata

    def __eq__(self, other):
        if isinstance(other, ManifestParameters):
            return (
                other.telescope.value == self.telescope.value
                and other.ngas_ingest == self.ngas_ingest
                and other.calibrate == self.calibrate
                and other.staging_source_dir == self.staging_source_dir
                and other.additional_metadata.filename == self.additional_metadata.filename
                and other.additional_metadata.content == self.additional_metadata.content
            )

        return False

    def to_dict(self) -> JSON:
        """
        form dictionary object for this manifest section

        :return:
        """
        json_dict = {
            ParamsKey.TELESCOPE.value: self.telescope.value,
            # The ingestion manifest must have "true" and "false"
            # rather than "True" and "False"
            ParamsKey.NGAS_INGEST.value: self.ngas_ingest,
        }
        if hasattr(self, "staging_source_dir"):
            json_dict[ParamsKey.INGESTION_PATH.value] = str(self.staging_source_dir)
        if hasattr(self, "calibrate"):
            json_dict[ParamsKey.CALIBRATE.value] = self.calibrate
        if self.additional_metadata:
            json_dict[ParamsKey.ADDITIONAL_METADATA.value] = self.additional_metadata.filename

        return json_dict


class AncillaryProduct(ManifestComponentIF):
    """Represents an ancillary product in an ingestion manifest"""

    def __init__(
        self,
        type: AncillaryProductType,
        filename: str,
        science_associate: str = None,
        group_with: str = None,
    ):
        self.type = type
        self.filename = filename
        self.science_associate = science_associate
        self.group_with = group_with

        # make this an ancillary to a particular science product (assumes locator string)
        if science_associate:
            self.science_associate = science_associate
        # make this an ancillary to the group of a science product (assumes locator string)
        if group_with:
            self.group_with = group_with

    def __eq__(self, other):
        if isinstance(other, AncillaryProduct) and other.type == self.type and other.filename == self.filename:
            return other.group_with == self.group_with and other.science_associate == self.science_associate

        return False

    def __str__(self):
        return f"{self.filename}: {self.type.value}"

    def to_dict(self) -> JSON:
        """
        Turn me into a json-ifiable dict

        :return: dicty-me
        """
        me_dict = dict(self.__dict__)
        me_dict["type"] = self.type.value
        clean_dict = {}
        for key, val in me_dict.items():
            if val is not None:
                clean_dict[key] = val

        return clean_dict


class OutputScienceProduct(ManifestComponentIF):
    """Generic science product contained in manifest output group"""

    def __init__(
        self,
        product_type: Union[ScienceProductType, AncillaryProductType],
        filename: str,
        ancillary_products: List[AncillaryProduct] = None,
    ):
        self.type = product_type
        self.filename = filename
        self.ancillary_products = ancillary_products

    def __eq__(self, other):
        if isinstance(other, OutputScienceProduct):
            return (
                other.type == self.type
                and other.filename == self.filename
                and other.ancillary_products == self.ancillary_products
            )

        return False

    def __str__(self):
        return f"{Path(self.filename).name}: {self.type.value}, {len(self.ancillary_products)} ancillary products"

    def to_dict(self) -> JSON:
        json_dict = {"type": self.type.value, "filename": self.filename}
        if self.ancillary_products:
            aps_json = [ap.to_dict() for ap in self.ancillary_products]
            json_dict[IngestionManifestKey.ANCILLARY_PRODUCTS.value] = aps_json
        return json_dict


class OutputGroup(ManifestComponentIF):
    """Generic ingestion manifest output group"""

    def __init__(
        self,
        science_products: List[OutputScienceProduct],
        ancillary_products: List[AncillaryProduct] = None,
    ):
        self.science_products = science_products
        self.ancillary_products = ancillary_products

    def __eq__(self, other):
        if isinstance(other, OutputGroup):
            return (
                other.science_products == self.science_products and other.ancillary_products == self.ancillary_products
            )

        return False

    def to_dict(self) -> JSON:
        """
        Turn me into a json-ifiable dict

        :return: dicty-me
        """
        me_dict = dict(self.__dict__)

        sps = me_dict[IngestionManifestKey.SCIENCE_PRODUCTS.value]
        sps = [sp.to_dict() for sp in sps]
        me_dict[IngestionManifestKey.SCIENCE_PRODUCTS.value] = sps

        aps = me_dict[IngestionManifestKey.ANCILLARY_PRODUCTS.value]
        if aps:
            ap_jsons = [ap.to_dict() for ap in aps]
            me_dict[IngestionManifestKey.ANCILLARY_PRODUCTS.value] = ap_jsons

        return me_dict


class ReingestGroup(ManifestComponentIF):
    """Generic manifest reingest group"""

    def __init__(
        self,
        locator: str,
        product_type: ScienceProductType,
        targets: List[str] = None,  # optional: presence determines Partial vs Full curation.
    ):
        self.locator = locator
        self.product_type = product_type
        self.targets = targets

    def __eq__(self, other):
        if isinstance(other, ReingestGroup):
            return (
                other.locator == self.locator
                and other.product_type == self.product_type
                and other.targets == self.targets
            )

        return False

    def to_dict(self) -> JSON:
        """
        Turn me into a json-ifiable dict

        :return: dict
        """
        json_dict = {
            ReingestKey.TARGETS.value: self.targets,
            ReingestKey.LOCATOR.value: self.locator,
            ReingestKey.TYPE.value: self.product_type.value,
        }

        return json_dict
